/*   File: LinkedList.h
 * Author: Socratis Tornaritis
 */

#ifndef LINKEDLIST_H
#define LINKEDLIST_H

#include <string>
#include <sstream>
#include <iostream>
#include <stdexcept>

using namespace std;


template <typename E>
class LinkedList {
private:
    // Forward class declaration.
    class Node;
    using NodePtr =  Node *;
    /*-------------------------------------------------------------------------------------*/

    /* The Node class is private to the clients of the LinkedList class, but its fields are
     * public, so the the methods of the LinkedList can access them directly. This avoids
     * the unnecessary use of accessor methods.
     */
    class Node {
    public:
        E element;
        NodePtr next;

        Node(const E& element, NodePtr const next): element{element}, next{next}
        {}// end Node()
    };
    /*-------------------------------------------------------------------------------------*/
    /* Utility methods. */
    NodePtr createNode(const E& element, NodePtr const next) {
        try {
            return new Node{element, next};

        } catch (bad_alloc e) {
            /* We are handling it here since if no memory then no point in moving on. */
            cerr << "Memory error: " << e.what() << endl;
            exit(1);
        }
    }// end createNode()
    /*-------------------------------------------------------------------------------------*/

    void insertNodeAfter(NodePtr const before, const E& element) {
        /* Insert a new node between 'before' and 'after'. */
        if ( isEmpty() ) {
            head = tail = createNode(element, nullptr);

        } else if ( before == nullptr ) {
            head = createNode(element, head);

        } else {
            NodePtr after = before->next;
            before->next = createNode(element, after);
        }
        count += 1;
    }// end insertNodeAfter()

    E removeNodeAfter(NodePtr const before) {
        /* Remove the 'del' node from the list.
         * Capture and return the element removed.
         */
        E elementRemoved;
        NodePtr del = nullptr;

        if ( before == nullptr ) {
            del = head;
            elementRemoved = del->element;
            head = head->next;

        } else {
            del = before->next;
            elementRemoved = del->element;
            NodePtr after = del->next;
            before->next = after;
        }

        if ( del == tail ) {
            tail = before;
        }
        /* Now, release the node itself. */
        releaseNode(del);
        count -= 1;
        return elementRemoved;
    }// end removeNode()

    void releaseNode(NodePtr& del) {
        /* Release the memory held by the node. */
        del->next = nullptr;
        delete del;
        del = nullptr;
    }// end releaseNode()
    /*-------------------------------------------------------------------------------------*/

    void copyNodes(const LinkedList& copy) {
        /* Method copies all nodes upon creation of a LinkedList object. */
        count = copy.count;

        /* Duplicate the first node on the copy list. */
        NodePtr iter = copy.head;

        head = tail = createNode(iter->element, nullptr);
        iter = iter->next;

        /* Visit each node on the copy list and duplicate it. */
        while ( iter != nullptr ) {
            tail = tail->next = createNode(iter->element, nullptr);
            iter = iter->next;
        }
    }// end copyNodes()
    /*-------------------------------------------------------------------------------------*/

    void releaseNodes() {
        NodePtr del = head;

        while ( del != nullptr ) {
            head = head->next;

            releaseNode(del);
            del = head;
        }
        tail = nullptr;
        count = 0;
    }// end releaseNode()
    /*-------------------------------------------------------------------------------------*/

    /* Fields. */
    NodePtr head = nullptr;
    NodePtr tail = nullptr;
    int count = 0;
    /*-------------------------------------------------------------------------------------*/

public:
    LinkedList() {}// end LinkedList()
    /*-------------------------------------------------------------------------------------*/

    LinkedList(const LinkedList& copy) { copyNodes(copy); }// end LinkedList()
    ~LinkedList()                      { releaseNodes();  }// end ~LinkedList()
    /*-------------------------------------------------------------------------------------*/

    const LinkedList& operator =(const LinkedList& rhs) {
        if ( this != &rhs) {
            releaseNodes();
            copyNodes(rhs);
        }

        return *this;
    }// end operator =()
    /*-------------------------------------------------------------------------------------*/
    /* Position-based methods */
    
    E get(const int index) const { 
        /* Get the element of some node within the linked list. */
        if ( isEmpty() ) { 
            throw logic_error("List is empty");                       
        }

        if ( index < 0 || index >= count ) { 
            throw logic_error("Invalid index: " + to_string(index) ); 
        }

        if ( index == 0 ) {
            return head->element;

        } else if ( index == count - 1 ) {
            return tail->element;

        } else {
            NodePtr iter = head;
            for (int move = 1; move <= index; move++) {
                iter = iter->next;
            }

            return iter->element;
        }
    }// end get()


    void set(const int index, const E& element) const { 
        /* Update the element of some node within the linked list. */
        if ( index < 0 || index >= count ) { 
            throw logic_error("Invalid index: " + to_string(index) ); 
        }

        if ( index == 0 ) {
            head->element = element;

        } else if ( index == count - 1 ) {
            tail->element = element;

        } else {
            NodePtr iter = head;
            for (int move = 1; move <= index; move++) {
                iter = iter->next;
            }

            iter->element = element;
        }
    }// end set()
    /*-------------------------------------------------------------------------------------*/
    

    void add(const int index, const E& element) {
        if ( index < 0 || index > count ) { 
            throw range_error( "Invalid index: " + to_string(index) ); 
        }

        if ( index == 0 ) { 
            insertNodeAfter(nullptr, element);

        }  else if ( index == count ) { 
            insertNodeAfter(tail, element);       

        } else {
            /* Insert within. */
            NodePtr before = head;
            
            for (int move = 1; move < index; move++) {
                before = before->next;
            }

            /* Insert the new node. */
            insertNodeAfter(before, element);
        }
    }// end add()
    /*-------------------------------------------------------------------------------------*/


    E removeAt(const int index) {
        if ( isEmpty() ) { 
            throw logic_error("List is empty");                        
        }

        if ( index < 0 || index >= count ) { 
            throw logic_error( "Invalid index: " + to_string(index) ); 
        }

        if ( index == 0 ) { 
            return removeNodeAfter(nullptr);

        } else {
            /* Remove from within. */
            NodePtr before = head;

            for (int move = 1; move < index; move++) {
                before = before->next;
            }
            return removeNodeAfter(before);
        }
    }// end popAt()
    /*-------------------------------------------------------------------------------------*/

    /* Value-based methods. */
    void add(const E& element) {
        /* This insertion is based on the value, not the index. 
         * The node will be inserted in sorted order, so this method should
         * not be used along with the index based methods, since that could
         * interfere with the sorted order.
         */
        
        if ( isEmpty() ) {
            add(0, element);

        } else if ( element < head->element ) {
            add(0, element);

        } else if ( element > tail->element ) {
            /* Insert to the back of the list. */
            add(count, element);

        } else {
            NodePtr before = nullptr;
            NodePtr after = head;

            while ( element > after->element ) {
                before = after;
                after = after->next;
            }

            insertNodeAfter(before, element);
        }
     }// end push()
    /*-------------------------------------------------------------------------------------*/

    void remove(const E& element) {  
        if ( isEmpty() ) { throw logic_error("List is empty"); }
 
        if ( element == head->element ) {
            removeAt(0);

        } else {
            /* The element must be located now within the sorted list. */
            NodePtr before = head;
            for ( NodePtr del = head->next; del != nullptr; del = del->next) {
                if ( element == del->element) {
                    removeNodeAfter(before);
                    break;

                } else if ( del->element > element ) {
                    ostringstream oss(element);
                    throw logic_error( "Element not found: " + oss.str() );
                }
                before = del;
            }
        }
    }// end pop()
    /*-------------------------------------------------------------------------------------*/

    /* Misc methods */
    bool isEmpty() const { return count == 0; }// end isEmpty()
    int size() const     { return count;      }// end size()
    /*-------------------------------------------------------------------------------------*/

    bool contains(const E& element) const {
        bool elementFound = false;
        NodePtr iter = head;

        while ( iter != nullptr && iter->element != element ) {
            iter = iter->next;
        }

        if ( iter != nullptr ) {
            elementFound = true;
        }

        return elementFound;
    }// end contains()
    /*-------------------------------------------------------------------------------------*/
    
    string toString() const {
        ostringstream oss;

        if ( isEmpty() ) { return "[ ]"; } 
        else {
            NodePtr iter = head;
            oss << "[ ";
            while ( iter != nullptr ) {
                oss << iter->element << ", ";
                iter = iter->next;
            }
            // Remove the last comma and space.
            oss.seekp(-2, ios::end);

            // Add the space back now.
            oss << " ]";
        }

        return oss.str();
    }// end toString()
/*-------------------------------------------------------------------------------------*/

}; // end LinkedList
/*-------------------------------------------------------------------------------------*/

template <typename E>
ostream& operator <<(ostream& os, const LinkedList<E>& o) {
    os << o.toString();
    return os;
}// end operator <<()
/*-------------------------------------------------------------------------------------*/

#endif