/*   File: LinkedList.h
 * Author: Socratis Tornaritis
 */

#ifndef LINKEDLIST_H
#define LINKEDLIST_H

#include <string>
#include <sstream>
#include <iostream>
#include <stdexcept>

using namespace std;


template <typename E>
class LinkedList {
private:
    class Node;
    using NodePtr =  Node *;
    /*-------------------------------------------------------------------------------------*/

    /* The Node class is private to the clients of the LinkedList class, but its fields are
     * public, so the the methods of the LinkedList can access them directly. This avoids
     * the unnecessary use of accessor methods.
     */
    class Node {
    public:
        E element;
        NodePtr next;

        Node(const E& element, NodePtr const next)
            : element{element}, next{next}
        {}// end Node()
    };
    /*-------------------------------------------------------------------------------------*/
    /* Utility methods. */
    
    NodePtr createNode(const E& element, NodePtr const next) {
        try {
            return new Node(element, next);

        } catch (bad_alloc e) {
            cerr << "Memory error: " << e.what() << endl;
            exit(1);
        }
    }// end createNode()
    /*-------------------------------------------------------------------------------------*/

    void insertNodeAfter(NodePtr const before, const E& element) {
        /* Insert a new node between 'before' and 'after'.
         * As this is the low-level insert routine, it will account for all states of
         * the list.
        */
       
        NodePtr after = before->next;
        before->next = createNode(element, after);

        if ( after == nullptr ) { tail = tail->next; }
        count += 1;
    }// end insertNode()

    E removeNodeAfter(NodePtr const before) {
        /* Remove the 'del' node from the list.
         * Capture and return the element removed.
         * Account for all states of the list.
        */
       if ( isEmpty() ) { throw range_error("List is empty"); }

        NodePtr del = before->next;
        NodePtr after = del->next;
        E elementPopped = del->element;

        before->next = after;

        /* Now, release the node itself. */
        releaseNode(del);
        count -= 1;

        if ( after == nullptr ) { tail = before; }
        return elementPopped;
    }// end removeNode()

    void releaseNode(NodePtr& del) {
        /* Release the memory held by the node. */
        del->next = nullptr;
        delete del;
        del = nullptr;
    }// end releaseNode()
    /*-------------------------------------------------------------------------------------*/

    void copyNodes(const LinkedList& copy) {
        /* Method copies all nodes upon creation of a LinkedList object. */
        count = copy.count;

        /* Duplicate the first node on the copy list. */
        NodePtr iter = copy.head;

        head = tail = createNode(iter->element, nullptr);
        iter = iter->next;

        /* Visit each node on the copy list and duplicate it. */
        while ( iter != nullptr ) {
            tail = tail->next = createNode(iter->element, nullptr);
            iter = iter->next;
        }
    }// end copyNodes()
    /*-------------------------------------------------------------------------------------*/

    void releaseNodes() {
        NodePtr del = head;

        while ( del != nullptr ) {
            head = head->next;

            releaseNode(del);
            del = head;
        }
        tail = nullptr;
        count = 0;
    }// end releaseNode()
    /*-------------------------------------------------------------------------------------*/

    /* Fields. */
    NodePtr head = nullptr;
    NodePtr tail = nullptr;
    int count = 0;
    /*-------------------------------------------------------------------------------------*/

public:
    LinkedList() {
        /* The linked list uses a dummy head node. */ 
        head = tail = createNode(E{}, nullptr);
        count = 0;
    }// end LinkedList()
    /*-------------------------------------------------------------------------------------*/

    LinkedList(const LinkedList& copy) { copyNodes(copy); }// end LinkedList()
    ~LinkedList()                      { releaseNodes();  }// end ~LinkedList()
    /*-------------------------------------------------------------------------------------*/

    const LinkedList& operator =(const LinkedList& rhs) {
        if ( this != &rhs) {
            releaseNodes();
            copyNodes(rhs);
        }

        return *this;
    }// end operator =()
    /*-------------------------------------------------------------------------------------*/

    /* Position-based Get/Set methods */
    E getFront() const { 
        return getAt(0);
    }// end getFront()

    E getBack() const  { 
        return getAt(count - 1);
    }// end getBack()

    E getAt(const int index) const { 
        /* Get the element of some node within the linked list. */
        if ( isEmpty() )                   { throw range_error("List is empty");                       }
        if ( index < 0 || index >= count ) { throw range_error("Invalid index: " + to_string(index) ); }

        /* Check for retrievals at the ends. */
        if      ( index == 0 )        { return head->element; }
        else if ( index == count - 1) { return tail->element; }
        else {
            /* Retrieve within. */
            NodePtr iter = head->next;
            for (int move = 1; move <= index; move++) {
                iter = iter->next;
            }

            return iter->element;
        }
    }// end getAt()

    void setFront(const E& element) {
        setAt(0, element);
    }// end setFront()

    void setBack(const E& element) {
        setAt(count - 1, element);
    }// end setBack()

    void setAt(const int index, const E& element) {
        /* Set the element of some node within the linked list. */
        if ( isEmpty() )                   { throw range_error("List is empty");                       }
        if ( index < 0 || index >= count ) { throw range_error("Invalid index: " + to_string(index) ); }

        /* Check for updates at the ends. */
        if      ( index == 0 )        { head->element = element; }
        else if ( index == count - 1) { tail->element = element; }
        else {
            /* Set within. */
            NodePtr iter = head->next;
            for (int move = 1; move <= index; move++) {
                iter = iter->next;
            }

            iter->element = element;
        }
    }// end setAt()
    /*-------------------------------------------------------------------------------------*/
    /* Position-based methods. */

    void pushFront(const E& element) { 
        pushAt(0, element);
    }// end pushFront()

    void pushBack(const E& element)  { 
        pushAt(count, element);
    }// end pushBack()

    void pushAt(const int index, const E& element) {
        if ( index < 0 || index > count ) { 
            throw range_error( "Invalid index: " + to_string(index) ); 
        }

        if ( index == count ) { 
            insertNodeAfter(tail, element); 

        } else {
            /* Insert within. */
            NodePtr before = head;
            
            for (int move = 1; move <= index; move++) {
                before = before->next;
            }

            /* Insert the new node. */
            insertNodeAfter(before, element);
        }
    }// end pushAt()

    E popFront() { 
        return popAt(0);
    }// end popFront()

    E popBack() {
        return popAt(count - 1);
    }// end popBack()

    E popAt(const int index) {
        if ( index < 0 || index >= count ) { throw range_error( "Invalid index: " + to_string(index) ); }

        NodePtr before = head;

        for (int move = 1; move <= index; move++) {
            before = before->next;
        }

        /* Remove the node from the list */
        return removeNodeAfter(before);
    }// end popAt()
    /*-------------------------------------------------------------------------------------*/
    /* Value-based methods. */

    void push(const E& element) {
        /* This insertion is based on the value, not the index. 
         * The node will be inserted in sorted order, so this method should
         * not be used along with the index based methods, since that could
         * interfere with the sorted order.
         */
        NodePtr before = head;
        
        if ( isEmpty() ) {
            pushAt(0, element);

        } else if ( element > tail->element ) {
            /* Insert to the back of the list. */
            pushAt(count, element);
    
        } else {
            NodePtr after = before->next;
            while ( element > after->element ) {
                before = after;
                after = after->next;
            }

            insertNodeAfter(before, element);
        }
     }// end push()
    /*-------------------------------------------------------------------------------------*/

    E pop(const E& element) {  

        if ( isEmpty() ) { throw range_error("List is empty"); }

        NodePtr del = nullptr;
        NodePtr before = nullptr;

        for ( before = head; before != nullptr; before = before->next ) {
            del = before->next;

            if ( del == nullptr || del->element > element ) {
                ostringstream oss(element);
                throw range_error( "Element not found: " + oss.str() );

            } else if ( del->element == element) {
                break;
            }
        }
        return removeNodeAfter(before);
    }// end pop()
    /*-------------------------------------------------------------------------------------*/

    /* Misc methods */
    bool isEmpty() const { return count == 0; }// end isEmpty()
    int size() const     { return count;      }// end size()
    /*-------------------------------------------------------------------------------------*/

    bool contains(const E& element) const {
          /* Search until you a) find it or b) exhaust the list. */
        for (NodePtr iter = head->next; iter != nullptr; iter = iter->next) {
            if ( iter->element == element ) {
                return true;
            }
        }
        return false;
    }// end contains()
    /*-------------------------------------------------------------------------------------*/
    
    string toString() const {
        ostringstream oss;

        if ( isEmpty() ) { return "[ ]"; } 
        else {
            NodePtr iter = head->next;
            oss << "[ ";
            while ( iter != nullptr ) {
                oss << iter->element << ", ";
                iter = iter->next;
            }
            // Remove the last comma and space.
            oss.seekp(-2, ios::end);

            // Add the space back now.
            oss << " ]";
        }

        return oss.str();
    }// end toString()
/*-------------------------------------------------------------------------------------*/

}; // end LinkedList
/*-------------------------------------------------------------------------------------*/

template <typename E>
ostream& operator <<(ostream& os, const LinkedList<E>& o) {
    os << o.toString();
    return os;
}// end operator <<()
/*-------------------------------------------------------------------------------------*/

#endif