/*   File: LinkedList.h
 * Author: Socratis Tornaritis
 * 
 * This version has the following improvements.
 * 
 * Two utility methods have been added to manage insertion and removal of nodes.
 *   void insertNodeAfter(NodePtr const before, const E& element, NodePtr const after) - inserts a new node
 *            between 'before' and 'after'.
 * 
 *   E removeNode(NodePtr const del) - removes then releases the 'del' node.
 * The use of these two methods allows us to remove duplicate code and make the code cleaner, more efficient
 * and easier to write/read.
 * 
 * Exception handling was added to the createNode() method. Since failure to allocate memory does not
 * allow us to continue, we will handle this error locally and exit the application.
 */

#ifndef LINKEDLIST_H
#define LINKEDLIST_H

#include <string>
#include <sstream>
#include <iostream>
#include <stdexcept>

using namespace std;


/* This class will be a bi-directional linked list, using a head and tail dummy node. 
 * E represents the element's type.
 * 
 * The linked list will offer value-based methods as well as position based methods.
 * 
 *     head                                                   tail
 *      |                                                      |
 * (nullptr, X, next)<-->(previous, E, next)<-->...(previous, X, nullptr)
 */
template <typename E>
class LinkedList {
private:
    class Node;
    using NodePtr =  Node *;
    /*-------------------------------------------------------------------------------------*/

    /* The Node class is private to the clients of the LinkedList class, but its fields are
     * public, so the the methods of the LinkedList can access them directly. This avoids
     * the unnecessary use of accessor methods.
     */
    class Node {
    public:
        NodePtr previous;
        E element;
        NodePtr next;

        Node(NodePtr const previous, const E& element, NodePtr const next)
            : previous{previous}, element{element}, next{next}
        {}// end Node()
    };
    /*-------------------------------------------------------------------------------------*/
    /* Utility methods. */
    NodePtr createNode(NodePtr const previous, 
                       const E& element, 
                       NodePtr const next) {
        try {
            // return the node's address on success.
            return new Node{previous, element, next};

        } catch (bad_alloc e) {
            cerr << "Memory error: " << e.what() << endl;
            exit(1);
        }
    }// end createNode()
    /*-------------------------------------------------------------------------------------*/

    void insertNodeAfter(NodePtr const before, const E& element) {
        /* Insert a new node between 'before' and 'after'. */
        NodePtr after = before->next;
        before->next = after->previous = createNode(before, element, after);
        count+= 1;
    }// end insertNodeAfter()

    E removeNode(NodePtr& del) {
        /* Remove the 'del' node from the list.
         * Capture and return the element removed.
         */
        E elementPopped = del->element;

        NodePtr before = del->previous;
        NodePtr after = del->next;

        before->next = after;
        after->previous = before;

        /* Now, release the node itself. */
        releaseNode(del);
        count-= 1;

        return elementPopped;
    }// end removeNode()

    void releaseNode(NodePtr& del) {
        /* Release the memory held by the node. */
        del->previous = del->next = nullptr;
        delete del;

        // This reset of del will be seen by the argument, since we are passing it by reference.
        del = nullptr;
    }// end releaseNode()
    /*-------------------------------------------------------------------------------------*/

    void copyNodes(const LinkedList& copy) {
        /* Method copies all nodes upon creation of a LinkedList object. */
        count= copy.count;

        /* Duplicate the first node on the copy list. */
        NodePtr iter = copy.head;

        head = tail = createNode(nullptr, iter->element, nullptr);
        iter = iter->next;

        /* Visit each node on the copy list and duplicate it. */
        while ( iter != nullptr ) {
            tail = tail->next = createNode(tail, iter->element, nullptr);
            iter = iter->next;
        }
    }// end copyNodes()
    /*-------------------------------------------------------------------------------------*/

    void releaseNodes() {
        NodePtr del = head;

        while ( del != nullptr ) {
            head = head->next;

            releaseNode(del);
            del = head;
        }
        tail = nullptr;
        count= 0;
    }// end releaseNode()
    /*-------------------------------------------------------------------------------------*/

    /* Fields. */
    NodePtr head = nullptr;
    NodePtr tail = nullptr;
    int count= 0;
    /*-------------------------------------------------------------------------------------*/

public:
    LinkedList() {
        /* The linked list uses dummy nodes. */ 
        head = tail = createNode(nullptr, E{}, nullptr);
        tail = tail->next = createNode(head, E{}, nullptr);
        count = 0;
    }// end LinkedList()
    /*-------------------------------------------------------------------------------------*/

    LinkedList(const LinkedList& copy) { copyNodes(copy); }// end LinkedList()
    ~LinkedList()                      { releaseNodes();  }// end ~LinkedList()
    /*-------------------------------------------------------------------------------------*/

    const LinkedList& operator =(const LinkedList& rhs) {
        if ( this != &rhs) {
            releaseNodes();
            copyNodes(rhs);
        }

        return *this;
    }// end operator =()
    /*-------------------------------------------------------------------------------------*/

    /* Position-based Get methods */
    E getFront() const { 
        return getAt(0);
    }// end getFront()

    E getBack() const  { 
        return getAt(count - 1);
    }// end getBack()

    E getAt(const int index) const { 
        /* Get the element of some node within the linked list. */
        if ( isEmpty() )                  { throw logic_error("List is empty");                       }
        if ( index < 0 || index >= count) { throw logic_error("Invalid index: " + to_string(index) ); }

        NodePtr iter = head->next;

        /* Check for retrievals at the end. */
        if ( index == count- 1) { 
            iter = tail->previous; 

        } else {
            for (int move = 1; move <= index; move++) {
                iter = iter->next;
            }
        }
        return iter->element;
    }// end getAt()

    void setFront(const E& element) {
        setAt(0, element);
    }// end setFront()

    void setBack(const E& element) {
       setAt(count - 1, element);
    }// end setBack()

    void setAt(const int index, const E& element) {
        /* Set the element of some node within the linked list. */
        if ( isEmpty() )                   { throw range_error("List is empty");                       }
        if ( index < 0 || index >= count ) { throw range_error("Invalid index: " + to_string(index) ); }

        NodePtr iter = head->next;

        /* Check for updates at the ends. */
        if ( index == count - 1) { 
            iter = tail->previous;  

        } else {
            /* Set within. */
            for (int move = 1; move <= index; move++) {
                iter = iter->next;
            }
        }
        iter->element = element;
    }// end setAt()
    /*-------------------------------------------------------------------------------------*/

    /* Position-based methods. */

    void pushFront(const E& element) { 
        pushAt(0, element);
    }// end pushFront()

    void pushBack(const E& element)  { 
         pushAt(count, element);
    }// end pushBack()

    void pushAt(const int index, const E& element) {
        if ( index < 0 || index > count) { 
            throw range_error( "Invalid index: " + to_string(index) ); 
        }

        NodePtr before = head;
        NodePtr after = before->next;

        /* Check for insertions at the ends. */
        if ( index == count) { 
            before = tail->previous;
            after = tail;

         } else {
            for (int move = 1; move <= index; move++) {
                after = after->next;
            }
        }

        /* Insert the new node. */
        insertNodeAfter(before, element);
    }// end pushAt()
    /*-------------------------------------------------------------------------------------*/

    E popFront() { 
        return popAt(0);
    }// end popFront()
    
    E popBack()  { 
        return popAt(count - 1);
    }// end popBack()

    E popAt(const int index) {
        if ( isEmpty() )                  { throw logic_error("List is empty");                        } 
        if ( index < 0 || index >= count) { throw logic_error( "Invalid index: " + to_string(index) ); }

        NodePtr del = head->next;
      
        if ( index == count- 1 ) { 
            del = tail->previous;

        } else {
            for (int move = 1; move <= index; move++) {
                del = del->next;
            }
        }
        /* Remove the node from the list */
        return removeNode(del);
    }// end popAt()
    /*-------------------------------------------------------------------------------------*/

    /* Value-based methods. */
    void push(const E& element) {
        /* This insertion is based on the value, not the index. 
         * The node will be inserted in sorted order, so this method should
         * not be used along with the index based methods, since that could
         * interfere with the sorted order.
         */
        if ( isEmpty() ) {
            pushAt(0, element);

        } else if ( element > tail->previous->element ) {
            /* Insert to the back of the list. */
            pushAt(count, element);
    
        } else {
            NodePtr after = head->next;
            while ( element > after->element ) {
                after = after->next;
            }
            NodePtr before = after->previous;
            insertNodeAfter(before, element);
        }
     }// end push()
    /*-------------------------------------------------------------------------------------*/

    E pop(const E& element) {  
        if ( isEmpty() ) { throw logic_error("List is empty"); }

 
        if ( tail->previous->element == element ) {
            /* If the element is the last one */
            NodePtr del = tail->previous;
            return removeNode(del);

        } else {
            /* The element must be located now within the list.
             * Since this method can be used when the list is a position-based
             * list, we can't simply assume the elements are in order. Thus,
             * we must search. */
            for ( NodePtr del = head->next; del != tail; del = del->next ) {
                if ( del->element == element ) {
                    // Found it, so remove its node.
                    return removeNode(del);
                }
            }
            // Did not find it, so throw exception.
            ostringstream oss(element);
            throw logic_error( "Element not found: " + oss.str() );
        }
    }// end pop()
    /*-------------------------------------------------------------------------------------*/

    /* Misc methods */
    bool isEmpty() const { return count== 0; }// end isEmpty()
    int size() const     { return count;     }// end size()
    /*-------------------------------------------------------------------------------------*/

    bool contains(const E& element) const {
        bool elementFound = false;
        NodePtr iter = head->next;

        while ( iter != tail && iter->element != element ) {
            iter = iter->next;
        }

        if ( iter != tail ) {
            elementFound = true;
        }

        return elementFound;
    }// end contains()
    /*-------------------------------------------------------------------------------------*/
    
    string toString() const {
        ostringstream oss;

        if ( isEmpty() ) { return "[ ]"; } 
        else {
            NodePtr iter = head->next;
            oss << "[ ";
            while ( iter != tail ) {
                oss << iter->element << ", ";
                iter = iter->next;
            }
            // Remove the last comma and space.
            oss.seekp(-2, ios::end);

            // Add the space back now.
            oss << " ]";
        }

        return oss.str();
    }// end toString()
/*-------------------------------------------------------------------------------------*/

}; // end LinkedList
/*-------------------------------------------------------------------------------------*/

template <typename E>
ostream& operator <<(ostream& os, const LinkedList<E>& o) {
    os << o.toString();
    return os;
}// end operator <<()
/*-------------------------------------------------------------------------------------*/

#endif